/*
Parametric 3D printable Strandbeest
by Alex Matulich
May 2020

This design on Thingiverse: https://www.thingiverse.com/thing:4390692
*/


// ============= adjustable parameters =============

/* [Printer-related settings] */

// Printer output: start with testfit to test and adjust tolerances
print="PETGparts"; // [testfit, PETGparts, nonPETGparts, cap, crankshaft, assembly, kneefootassembly]

// line width from nozzle
linewid = 0.45;
// layer height
lineht = 0.2;
// cylinder snug fit tolerance for lock ring
tighttol = 0.12;
// cylinder fit tolerance for bearing diameter
loosetol = 0.28;
// very tight fit tolerance for shaft key
tightfittol = 0.0;
// extra layers of crankcase slack
crankcase_layertol = 2;
// extra layers of crankshaft slack
crankshaft_layertol = 2;

/* [Hubs and axles] */
//scale_pivots = false;   // true=scale all pivot parameters by lengthscale (NOT IMPLEMENTED)

// number of print layers for a "washer"
hubwasherlayers = 3;
// diameter of hubs around axles
hubdia = 9;
// diameter of all axle shaftsbones
shaftdia = 6;
// diameter of foot tip (at least line width)
footdia = 1.0;

/* [Bone dimensions] */
// bone width compared to hub size
bonewidthfrac = 0.62;
// thickness of outer bone piece (thinnest)
outerboneht = 2.0;
// thickness of inner bone piece (thickest)
innerboneht = 4.6;
// thickness of crankcase (center frame)
crankcaseht = 5.0;
// thickness of crankshaft arm
crankarmht = 3.0;

/* [Other] */

// scale factor for bone dimension (length, width, height)
lengthscale = 1.0;
// number of facets on cylinders
fn = 40;    // [24:8:64]
// number of leg pairs
ncranks = 5;    // [1:6]
// text labels: all, or none except knee and foot
labels = true;

// ============= end of adjustable parameters =============

// having a module here forces the end of Customizer parameters

module lego_axle_hole(len) {
    ldia = 4.78 + loosetol;
    vwid = 1.79 + loosetol;
    translate([0,0,0.5*len]) {
        cube([ldia,vwid,len],true);
        cube([vwid,ldia,len],true);
    }
}

// ============= advanced parameters =============

crank_angle = [ // Crank angles for different size stacks of cranks. First angle should the first offset from zero. Second angle should be the offset from the first, and so on.
    [ 0 ],                        // unused (offset for zero-index of list
    [ 0 ],                        // 1 crank
    [ 180, 180 ],                 // 2 cranks
    [ 120, 120, 120 ],            // 3 cranks
    [ 90, 90, 90, 90 ],           // 4 cranks
    [ 142, 142, 142, 142, 142 ],  // 5 cranks
    [ 120, 120, -60, 120, 120, 120 ] // angles for stack of 6 cranks
];

accumangle = // accumulated crank angles
    [ for (n=[0:ncranks-1]) accumsum(crank_angle[ncranks], n) % 360 ];

/*
Strandbeest leg proportions from
https://commons.wikimedia.org/wiki/File:Strandbeest_Leg_Proportions.svg
body pivot is at intersection of B,C,D
crankshaft M rotates the intersection J,K around the top point of interval L
triangle BDE = knee
triangle GHI = foot
/ = narrow (inner) bone
// = wide (outer) bone with nested inner bone
/// = wider outer bone with nested wide and narrow bones

                    /\\\
                  / |  \\\
                /   |    \\\
              /     |      \\\
            /       |        \\\
          / E       |        J \\\
        /         B |            \\\
      /             |              \\\
    /               |          ._____\\\
  /                 |         L|  M   //
/___________________| . . . . .|    //
\\         D       \\\     A      //
 \\                 \\\         //
  \\ F             C \\\    K //
   \\                 \\\   //
    \\_________________\\\//
      \        G         /
        \               /
          \           I /
            \ H        /
              \        /
                \     /
                  \   /
                    \/
*/

// Theo Jansen's linkage lengths (per figure above)
// from https://en.wikipedia.org/wiki/Jansen%27s_linkage
///*
sa = 38.0 * lengthscale; // interval
sb = 41.5 * lengthscale; // single (part of knee)
sc = 39.3 * lengthscale; // double
sd = 40.1 * lengthscale; // single (part of knee)
se = 55.8 * lengthscale; // single (part of knee) - optional if b-d connection is strong
sf = 39.4 * lengthscale; // double
sg = 36.7 * lengthscale; // single (part of foot)
sh = 65.7 * lengthscale; // single (part of foot) - optional if g-i connection is strong
si = 49.0 * lengthscale; // single (part of foot)
sj = 50.0 * lengthscale; // double
sk = 61.9 * lengthscale; // triple (2 bones must nest inside)
sl =  7.8 * lengthscale; // interval
sm = 15.0 * lengthscale; // single (cranskaft diameter)
he_bend = 5; // how much to bend segments H and E inward (suggest 0-6) - this gets scaled by lengthscale
//*/
// Alternate linkage lengths for optimally flat ground contact path without sacrificing foot lift height
// from http://scottburns.us/walking-mechanism/
/*
ascl = 3.955; // convert to same scale as Jansen's numbers above
sa = 10*cos(16.1) * ascl*lengthscale; // segment A is the reference
sb = 12.9 * ascl*lengthscale;
sc = 9.4 * ascl*lengthscale;
sd = 10.0 * ascl*lengthscale;
se =  8.3 * ascl*lengthscale;
sf = 10.1 * ascl*lengthscale;
sg =  9.4 * ascl*lengthscale;
sh = 16.8 * ascl*lengthscale;
si = 11.6 * ascl*lengthscale;
sj = 18.5 * ascl*lengthscale;
sk = 15.7 * ascl*lengthscale;
sl = 10*sin(16.1) * ascl*lengthscale;
sm =  3.1 * ascl*lengthscale;
he_bend = 1.32*ascl;
*/

// ============= initialization =============

// lock ring and lock shaft values

lockshaftd=0.8*shaftdia; // locking shaft radius
lockringwid = 0.5*(shaftdia-lockshaftd); // width of lock ring
lockrimht = lineht*lengthscale; // height of rim in middle of lock ring
lockringht = lockrimht + 2*lockringwid; // total height of lock ring
lockringgap = 1.5*lockringwid; // gap width to cut out of shaft

lockshafthole_dt = lockshaftd-2.5 > 0 ? lockshaftd-2.5 : 0.2; // top of hole to cut out of shaft
lockshafthole_db2 = shaftdia-3 > 0 ? shaftdia-2.6 : 0.2; // top of bottom hole
lockshafthole_db1 = shaftdia-3 > 0 ? shaftdia-3.4 : 0.2; // bottom of hole
lockshafthole_th = lockshafthole_db2-lockshafthole_dt;
lockneckht = 2*lineht; // length of neck (end of shaft to bottom of lock ring
minlockht = 2*lockringwid+lockrimht+lockneckht; // total height of locking mechanism

// bone height constraints

hubwasher = hubwasherlayers * lineht * lengthscale;
obht = max(minlockht, outerboneht * lengthscale);
if (obht > outerboneht*lengthscale) echo("OUTERBONE TOO THIN; INCREASED TO:", obht);
ibht = (max(innerboneht*lengthscale, 2*obht+hubwasher));
if (innerboneht < ibht) echo("INNERBONE TOO THIN; INCREASED TO:",ibht);
lockgapdepth = 1.1*shaftdia;

// other bone values

she_bend = he_bend * lengthscale;   // segments H&E bend amount
sk_bend = 0.5*hubdia;               // bone K bend amount
bonewid = hubdia * bonewidthfrac;   // bone width

sh1 = ibht + 2*hubwasher + lineht;  // level 1 shaft length
sh2 = ibht + 2*obht+4*hubwasher + lineht; // level 2 shaft length
brotate = 90-vertex_angle(sd,sb,se); // angle of bone b from vertical

// crankcase-related values

ccht = crankcaseht * lengthscale;   // crankcase thickness
caht = crankarmht * lengthscale;    // crank arm thickness

innerspacer = ibht - 2*obht;
crankshaftht =          // crankshaft holds a stack of:
    8*obht              // 8 outer bone pieces
    + innerspacer       // a spacer (2 bone K hub washers)
    + 8*hubwasher       // all the other washer/spacers
    + crankshaft_layertol*lineht; // a bit of slack

shaftkeywid = shaftdia*cos(45);
crankspace =            // total inner crankcase width holds:
    2*caht              // 2 crank arms
    + 2*hubwasher       // 2 hub washers
    + crankshaftht      // length of crankshaft
    + crankcase_layertol*lineht; // a bit more slack
handleht = caht;        // height of shaft end handle

// text placement

textr = hubdia * 0.6;
textht = obht-lineht;


// ============= print outputs =============

/*
Test the fit of some representative parts.
You can use these later or discard them if they will confuse you later.
The parts printed are:
1. Crankshaft pieces for SECOND stage crankcase, to test tight and loose crankshaft keys (the keys on the first stage are both tight).
2. Bone C, which includes two snap locks and a bearing, which should slide easily over any of the shafts.
3. A spacer used in the crankcase, which should slide easily over any of the shafts.
*/
if (print == "testfit") {
    bone_c(1);
    dy = hubdia+2;
    translate([0,dy,0]) bone_c(2);
    translate([0,2*dy,0]) crankshaft(2,1);
    translate([sm+2+hubdia, 2*dy, 0]) crankshaft(1,1);
    translate([-hubdia-2,0,0]) spacer(caht+2*hubwasher);
}

/*
Show all the bones for one leg, including knee and foot.
If you want some a different color, delete them in your slicer and print them separately.
*/
if (print == "allbones") {
    foot();
    translate([0, si+2.6*hubdia, 0]) knee();
    translate([sg+hubdia+2,0,0]) bone_c(1);
    translate([sg+hubdia+2,hubdia+2,0]) bone_c(2);
    translate([sg*0.8,2*(hubdia+2),0]) bone_j(1);
    translate([sg*0.8,3*(hubdia+2),0]) bone_j(2);
    translate([sg*0.6,4*(hubdia+2)+sk_bend,0]) bone_k(1);
    translate([sg*0.6+sk,5*(hubdia+2)+sk_bend+2,0]) rotate([0,0,180]) bone_k(2);
    translate([sd+hubdia+2,7*(hubdia+2),0]) bone_f(1);
    translate([sd+hubdia+2,8*(hubdia+2),0]) bone_f(2);
}

if (print == "PETGparts") {
    translate([-sc-2*hubdia, -4*hubdia,0]) bone_c(1);
    translate([-sc-2*hubdia, -2.5*hubdia,0]) bone_c(1);
    translate([-sj-1.5*hubdia, 1.5*hubdia,0]) bone_j(1);
    translate([-sj-1.5*hubdia, 3*hubdia,0]) bone_j(1);
    translate([1.5*hubdia, 3*hubdia,0]) bone_k(1);
    translate([1.5*hubdia, 1.5*hubdia,0]) bone_k(1);
    translate([2*hubdia, -2.5*hubdia,0]) bone_f(1);
    translate([2*hubdia, -4*hubdia,0]) bone_f(1);
    crankcase(showspacers=false);
}

if (print == "nonPETGparts") {
    translate([-sc-2*hubdia, -2.5*hubdia,0]) bone_c(2);
    translate([-sc-2*hubdia, -hubdia,0]) bone_c(2);
    translate([-sj-1.5*hubdia, hubdia,0]) bone_j(2);
    translate([0,hubdia,0]) spacer(hubht1);
    translate([-sj-1.5*hubdia, 2.5*hubdia,0]) bone_j(2);
    translate([0,2.5*hubdia,0]) spacer(hubht2);
    translate([-si-2*hubdia,sg+4*hubdia,0]) rotate([0,0,-85]) foot();
    translate([-hubdia,4*hubdia,0]) knee();
    translate([hubdia,4*hubdia,0]) rotate([0,0,-90]) knee();
    translate([sb+2*hubdia, si+3.5*hubdia, 0]) rotate([0,0,180]) foot();
    translate([1.5*hubdia, 2.5*hubdia,0]) bone_k(2);
    translate([1.5*hubdia, hubdia,0]) bone_k(2);
    translate([2*hubdia, -hubdia,0]) bone_f(2);
    translate([2*hubdia, -2.5*hubdia,0]) bone_f(2);
}

if (print == "crankcase") crankcase();

if (print == "cap") crankcase(cap=true);

if (print == "crankshaft") {
    translate([-1.5*hubdia-2,0,0]) crankshaft_end();
    for(i=[1:ncranks], yoff = (i-1)*(hubdia+2)) {
        translate([0, yoff, 0]) crankshaft(1,i-1);
        translate([hubdia+2*sm+2, yoff, 0]) rotate([0,0,180]) crankshaft(2,i-1);
        translate([2*(hubdia+sm+2), yoff, 0]) spacer(hubwasher);
    }
}

// ============= individual component modules =============

// build the knee (top triangular bone) from three joined bones

module knee() {
    //     /|
    //  e/  |b
    // /__d_|
    db = 90-brotate; //vertex_angle(sd,sb,se); // angle between segments d and b
    de = vertex_angle(sd,se,sb); // angle between segments d and e
    fillet = 0.2;
    v = triangle_vertices(sb,se,sd);
    textht = ibht-lineht;
    difference() {
        union() {
            translate([-sd,0,0]) bone(sd, bonewid, ibht, hubdia,hubdia, 0,0, joint1=true,joint2=true, bend=shaftdia/2);
            translate([se*cos(de)-sd,se*sin(de),0]) rotate([0,0,180+de]) bone(se, bonewid, ibht, hubdia,hubdia, 0,0, joint1=true, joint2=true, bend=she_bend);
            translate([0,0,0]) rotate([0,0,180-db]) bone(sb, bonewid, ibht, hubdia,hubdia, 0,0, joint1=true,joint2=true);
        }
        translate([-sd,0,0]) bearing(ibht);
        translate([0,0,0]) bearing(ibht);
        translate([v[2][0]-sd,v[2][1],0]) bearing(ibht);
        translate([-sd,0,0]) rotate([0,0,12]) translate([textr,0,textht]) linear_extrude(2*lineht, convexity=4) text(text="< F", size=shaftdia/2, font="Arial:style=Bold",halign="left", valign="center");
        translate([-textr*cos(db),textr*sin(db),textht]) rotate([0,0,180-db]) linear_extrude(2*lineht, convexity=4) text(text="< A", size=shaftdia/2, font="Arial:style=Bold",halign="left", valign="center");
        translate([v[2][0]+textr*cos(db)-sd,v[2][1]-textr*sin(db),textht]) rotate([0,0,180-db]) linear_extrude(2*lineht, convexity=4) text(text="J >", size=shaftdia/2, font="Arial:style=Bold",halign="right", valign="center");
    }
}


// build the foot (bottom triangular bone) from three joined bones

module foot() {
    gh = vertex_angle(sh,sg,si); // angle between segments g and h
    gi = vertex_angle(sg,si,sh); // angle between segments g and i
    v = triangle_vertices(sh,si,sg);
    textht = ibht-lineht;
    textr = hubdia*0.6;
    filletfrac = 0.25;

    fillet = [ // this is needed only if she_bend < 4;
        [v[2][0],v[2][1]],
        [ v[2][0]+filletfrac*sh*cos(gh+4), v[2][1]-filletfrac*sh*sin(gh+4) ],
        [ v[2][0]+filletfrac*si*cos(0.5*(180-gi+gh)), v[2][1]-filletfrac*si*sin(0.5*(180-gi+gh)) ],
        [ v[2][0]+filletfrac*sh*cos(180-gi-4), v[2][1]-filletfrac*sh*sin(180-gi-4) ]
    ];
    bw = (hubdia+footdia)/2;
    difference() {
        union() {
            bone(sg, bonewid, ibht, hubdia,hubdia, 0,0, joint1=true,joint2=true);
            translate([si*cos(gi),si*sin(gi),0]) rotate([0,0,180+gi]) bone(si, bw, ibht, footdia,hubdia, 0,0, joint1=true, joint2=true);
            translate([sg,0,0]) rotate([0,0,180-gh]) bone(sh, bw, ibht, hubdia,footdia, 0,0, joint1=true,joint2=true, bend=she_bend);
            if (she_bend < 4) linear_extrude(ibht, convexity=4) polygon(points=fillet);
        }
        bearing(ibht);
        translate([sg,0,0]) bearing(ibht);
        translate([textr,0,textht]) rotate([0,0,180]) linear_extrude(2*lineht, convexity=4) text(text="C >", size=shaftdia/2, font="Arial:style=Bold",halign="right", valign="center");
        translate([sg-textr,0,textht]) rotate([0,0,180]) linear_extrude(2*lineht, convexity=4) text(text="< F", size=shaftdia/2, font="Arial:style=Bold",halign="left", valign="center");
    }
}

// bones C, F, J, K -- parameter 'part' is 1 or 2, indicating which of two parts to generate

// bone C -- wide bone that connects crankcase to intersection of inner foot bearing and bottom of bone K
module bone_c(part) {
    sh_c = part == 1 ? sh2 : -1;
    bone(sc/2, bonewid, obht, hubdia,hubdia, hubwasher,0, sh_c,-2, label="C");
    translate([sc/2,0,0]) bone(sc/2, bonewid, obht, hubdia,hubdia, 0,obht+2*hubwasher, sh_c,-2, label="A >");
}

// bone F -- connects outer knee bearing to outer foot bearing
module bone_f(part) {
    sh_f = part == 1 ? sh1 : -1;
    bone(sf, bonewid, obht, hubdia,hubdia, hubwasher,hubwasher, sh_f,sh_f, label="F");
}

// bone J -- wide bone that connects crankshaft M1 to top of knee
module bone_j(part) {
    sh_j = part == 1 ? sh2 : -1;
    kneehub = obht + 2*hubwasher;
    difference() {
        union() {
            bone(sj-sm-shaftdia, bonewid, obht, hubdia,hubdia, kneehub,0, sh_j,-2, label="J");
            translate([sj-sm-shaftdia,0,0]) bone(sm+shaftdia, bonewid,obht, hubdia,hubdia, 0,hubwasher, sh_j,-2);
        }
        if (labels) translate([sj-textr,0,textht]) linear_extrude(2*lineht, convexity=4) text(text="M1 >", size=shaftdia/2, font="Arial:style=Bold",halign="right", valign="center");
    }
}

// bone K -- connects crankshaft M1 to intersection of inner foot bearing and bottom of bone C
module bone_k(part) {
    bend = (part == 1 ? sk_bend : -sk_bend);
    sh_k = part == 1 ? sh1 : -1;
    kp1 = 0.25*sc;  // x coord end of segment 1
    kp2 = sk/2;     // x coord end of segment 2
    hd2 = min(shaftdia+2, 0.5*(hubdia+shaftdia)); // hub dia end seg 2-needs to be a bit smaller for clearance
    sk1 = sqrt(1/16*sc*sc+sk_bend*sk_bend); // length of segment 1
    sk2 = sk - kp1 - kp2; // length of segment 2
    sk3 = sqrt(kp2*kp2+sk_bend*sk_bend); // length of segment 3
    sk2bend = part==1 ? -0.04*sk2 : 0.04*sk2;
    ak1 = atan(bend/kp1); // rotation angle of segment 1
    ak3 = atan(bend/kp2); // rotation angle of segment 3

    rotate([0,0,-ak1]) bone(sk1, 0.5*(hubdia+bonewid), obht, hubdia,hubdia, hubwasher,0, -2,-2);
    translate([kp1,-bend,0]) bone(sk2, bonewid, obht, hubdia,hd2, 0,0, sh_k,-2, bend=sk2bend, label="K");
    translate([sk-kp2,-bend,0]) rotate([0,0,ak3]) difference() {
        bone(sk3, bonewid, obht, hd2,hubdia, 0,part==1?innerspacer/2:hubwasher, sh_k,-2);

    if (labels) translate([sk3-textr,0,textht]) linear_extrude(2*lineht, convexity=4) text(text="M1 >", size=shaftdia/2, font="Arial:style=Bold",halign="right", valign="center");
    }
}

// build the crankcase
// When cap==true, an end-cap for the crankcase is generated

hubht1 = caht+2*hubwasher;
hubht2 = hubht1+2*(obht+hubwasher);
module crankcase(cap=false, showspacers=true) {
    alen = sqrt(sa*sa+sl*sl);
    ang = atan(sl/sa);
    hht1 = cap ? 0 : hubht1;
    hht2 = cap ? 0 : hubht2;
    vlen = sm+hubdia+0.5*hubdia;
    crankhubdia = 1.4*hubdia;
    bw = 0.5*(crankhubdia+hubdia)*bonewidthfrac;
    bend = 1*lengthscale;
    difference() {
        union() {
            rotate([0,0,-ang]) bone(alen, bw, ccht, crankhubdia, hubdia, 0,hht1, shaft2ht=cap?0:crankspace, bend=bend, label=cap?"":"A");
            translate([-alen*cos(ang),-alen*sin(ang),0]) rotate([0,0,ang]) bone(alen, bw, ccht, hubdia, crankhubdia, hht2,0, shaft1ht=cap?0:crankspace, bend=bend, label=cap?"":"A");
            rotate([0,0,-90]) bone(sa, bw, ccht, crankhubdia, hubdia, 0,0, shaft2ht=cap?0:crankspace);
            rotate([0,0,90]) bone(vlen, bw, ccht, crankhubdia, hubdia, 0,0, shaft2ht=cap?0:crankspace);
            if (!cap) hub(ccht, hubwasher, hubdia);
        }
        bearing(ccht);
        translate([-sa,-sl,0]) lockringhole(ccht,true,false);
        translate([sa,-sl,0]) lockringhole(ccht,true, false);
        translate([0,vlen,0]) lockringhole(ccht,true,false);
        translate([0,-alen,0]) lockringhole(ccht,true,false);
        if (labels && !cap) translate([-hubdia/2*cos(ang),-hubdia/2*sin(ang),lineht]) rotate([0,180,ang/2]) linear_extrude(2*lineht, convexity=4) text(text="< M2", size=shaftdia/2, font="Arial:style=Bold",halign="left", valign="center");
    }
    // make spacers
    if (!cap && showspacers) {
        translate([-sa/2,-vlen,0]) spacer(hubht1);
        translate([sa/2,-vlen,0]) spacer(hubht2);
    }
}

// The crankshaft has 3 components: 1=crank, 2=crank end, and washer. The washer is generated in the "crankshaft" print setting.

// Build half of a crankshaft.
// Parameters:
// which_half: 1=main crank arm and shaft (tight fit), 2=crank arm transition to next crankcase (loose fit)
// angle_index: 0,1,2... the index of crank_angle[ncranks] to use for setting the shaft key angle
module crankshaft(which_half, angle_index) {
    difference() {
        if (which_half==1)
            bone(sm, bonewid, caht, hubdia,hubdia, 0,hubwasher, -1, crankshaftht, shaftend=angle_index==0?3:2, keyang=0);
        else
            translate([sm,0,0]) rotate([0,0,180]) bone(sm, bonewid, caht, hubdia,hubdia, 0,hubwasher, -1, ccht+2*hubwasher+lineht, shaftend=angle_index==0?3:2, keyang=crank_angle[ncranks][angle_index]);

        // labels
        if (labels) translate([sm/2,0,caht-lineht]) linear_extrude(2*lineht, convexity=4) text(text=which_half==1?"M1":"M2", size=shaftdia/2, font="Arial:style=Bold",halign="center", valign="center");
        translate([sm/2,0,lineht-1]) linear_extrude(1, convexity=4) rotate([0,180,0]) text(text=str(angle_index+1), size=shaftdia/2, font="Arial:style=Bold",halign="center", valign="center");
    }
}

// Generate the zero-end of the crankshaft, to lock it into the first crankcase. We'll put a Lego axle hole in it.
module crankshaft_end() {
    difference() {
        union() {
            hub(handleht, 0, 2*hubdia);
            translate([0,0,handleht-0.01]) {
                hub(caht, hubwasher, hubdia);
                shaft(caht, ccht+2*hubwasher+crankshaft_layertol*lineht, shaftend=3, keyrot=0, keytol = tighttol); // keytol expands key
            }
        }
        translate([0,0,-1]) lego_axle_hole(handleht+caht/2);
    }
}

// ============= helper modules used to construct components =============

// bone module - used to make all "bones" including the knee and foot
// parameters:
// length:   length of bone between the pivots
// width:    width of bone at narrowest point
// ht:       height (thickness of bone)
// hub1ht:   height of hub 1; typically bone thickness+washer, ht used for any value < ht
// hub2ht:   height of hub 2; similar to hub 1
// shaft1ht: height of axle shaft 1; -2=hub hole, -1=shaft lock hole, 0=disable, else make shaft
// shaft2ht: height of axle shaft 2; same values as shaftht1
// label:    text to emboss for identification of part
// joint1:   true=flare top curve at hub 1 for joining with another bone; used by knee() and foot()
// joint2:   true=flare top curve at hub 2
// bend:     units of bend displacement at middle (1 = positive y bend, -1 = negative y bend)
// shaftend: 0=none, 1=snap lock, 2=loose crank key, 3=tight crank key
// keyang:   shaft key angle if shaftend==2
// bzfn:     number of bezier points to generate for each 1/4 of bone perimeter excluding end hubs
module bone(length, width, ht, d1, d2, hub1ht, hub2ht, shaft1ht=0, shaft2ht=0, label="", joint1=false, joint2=false, bend=0, shaftend=1, keyang=0, bzfn=16) {
    r1 = 0.5*d1;
    r2 = 0.5*d2;
    w = 0.5*width;
    maxht = ht+max([hub1ht,hub2ht]);
    keytol = shaftend==3 ? tightfittol : tighttol;

    // bezier control points, clockwise from origin
    bzt1 = [ joint1?[0.8*r1,1.2*r1]:[0,r1], [r1,r1], [0.15*length,w+bend], [0.5*length,w+bend] ]; // left top quarter
    bzt2 = [ [0.5*length,w+bend], [0.85*length,w+bend], [length-r2,r2], joint2?[length-0.8*r2,1.2*r2]:[length,r2] ]; // right top quarter
    bzb1 = [ [length,-r2], [length-r2,-r2], [0.85*length,-w+bend], [0.5*length,-w+bend] ]; // right bottom quarter
    bzb2 = [ [0.5*length,-w+bend], [0.15*length,-w+bend], [r1,-r1], [0,-r1] ];
    stp = 2/fn;
    eps = 0.01*stp;

    difference() {
        union() {
            // bezier curves
            linear_extrude(ht, convexity=4) polygon(points= [
                for(t=[0:stp:1-eps]) bezier_point2d(t, bzt1),
                for(t=[0:stp:1+eps]) bezier_point2d(t, bzt2),
                for(t=[0:stp:1-eps]) bezier_point2d(t, bzb1),
                for(t=[0:stp:1+eps]) bezier_point2d(t, bzb2)
            ]);
        
            // hub ends
            hub(ht, hub1ht, d1, hub1ht>0);
            translate([length,0,0]) hub(ht, hub2ht, d2, hub2ht>0);
    
            // shafts if any
            if (shaft1ht > 0) shaft(ht, shaft1ht, shaftend);
            if (shaft2ht > 0) translate([length,0,0]) shaft(ht, shaft2ht, shaftend, keyang);
        }
        // subract holes, if any
        if (shaft1ht < -1) bearing(maxht);
        else if (shaft1ht < 0) {
            if (shaftend==1) lockringhole(ht);
            else if (shaftend>=2) keyhole(ht, shaftend);
        }
        if (shaft2ht < -1) translate([length,0,0]) bearing(maxht);
        else if (shaft2ht < 0) {
            if (shaftend==1) translate([length,0,0]) lockringhole(ht);
            else if (shaftend>=2) translate([length,0,0]) keyhole(ht, shaftend);
        }
        // label
        if (labels) translate([d1,bend*2*d1/length,ht-lineht]) linear_extrude(2*lineht, convexity=4) text(text=label, size=shaftdia/2, font="Arial:style=Bold",halign="left", valign="center");
    }
}

// hub (a cylinder extending above a bone surface)
// parameters:
// boneht: thickness of bone on which the hub is built
// hubht: thickness of hub on the bone
// washer: true=make hub look like a narrow washer rather than a disc
module hub(boneht, hubht, dia, washer=true) {
    difference() {
        cylinder(h=boneht+hubht, d=dia, $fn=fn);
        if (washer) translate([0,0,boneht-0.01]) cylinder(h=hubht+1, d=shaftdia+2*linewid, $fn=fn);
    }
}

// bearing (a hole in a hub); this is used as a difference() object
// parameter:
// ht: thickness of object through which to cut the bearing hole
module bearing(ht) {
    translate([0,0,-1]) cylinder(h=ht+2, d=shaftdia+loosetol, $fn=fn);
}

// shaft
// parameters:
// boneht: thickness of bone under the shaft
// ht: height of shaft above the bone
// shaftend: 0=none, 1=snap fitting, 2,3=cranshaft key
// keyrot: rotation angle of crankshaft key
module shaft(boneht, ht, shaftend=1, keyrot=180, keytol=0) {
    top = ht+boneht;
    keydisp = -0.5*shaftkeywid;
    keyht = min(boneht,shaftkeywid);
    difference() {
        union() {
            cylinder(h=top, d=shaftdia, $fn=fn);
            if (shaftend==1) {
                translate([0,0,top-0.1]) cylinder(h=lockneckht+0.2, d=lockshaftd, $fn=fn); // neck
                lockring(top); // lock ring
            } else if (shaftend>=2)
                rotate([0,0,180+keyrot]) translate([0,0,top-0.01]) key(keyht,keytol);
        }
        if (shaftend==1) shaftcutout(top);
    }
}

// lockring -- create a locking ring
// parameter:
// shafttop: z-height of vertical shaft; if zero, it is assumed this is being subtracted to make a hollow space to fit the locking ring. The locking ring diameter is reduced by a tolerance and the hollow ring is increased by the same tolerance.
ringext = 1*lineht;
module lockring(shafttop) { // if shafttop==0, assume this is being used to make a hole
    tol = shafttop == 0 ? tighttol : -tighttol;
    d = shaftdia + tol;
    coneht = lockringwid + ringext;
    d2 = d - 2*coneht;
    z = shafttop == 0 ? 0 : shafttop-coneht+(d-lockshaftd)/2+lockneckht;
    union() {
        translate([0,0,z]) cylinder(h=coneht, d1=d2, d2=d, $fn=fn);
        translate([0,0,z+coneht]) cylinder(h=lockrimht, d=d, $fn=fn);
        translate([0,0,z+coneht+lockrimht]) cylinder(h=coneht, d1=d, d2=d2, $fn=fn);
    }
}

// shaftcutout -- cut out channels in the shaft so the locking ring can compress and snap into the locking hole
// parameter:
// ht: z-height of shaft top excluding the lock ring assembly
module shaftcutout(ht) {
    l = 1.5*shaftdia;
    //w = lockringgap;
    //r = 0.5*lockringgap;
    w1 = lockringgap*0.65;
    w2 = lockringgap*1.3;
    r = 0.5*w1;
    //h = lockgapdepth*2;
    h0 = lockgapdepth+lockringht+lockneckht;
    bot = ht-lockgapdepth+r;
    union() {
//        translate([-l/2,-w/2,bot]) cube([l,w,h]);
//        translate([-w/2,-l/2,bot]) cube([w,l,h]);
        translate([0,0,bot]) linear_extrude(h0, scale=w2/w1) square([l,w1], center=true); 
        translate([0,0,bot]) linear_extrude(h0, scale=w2/w1) square([w1,l], center=true); 

        translate([0,0,ht-0.1]) cylinder(h=minlockht+1, d=lockshafthole_dt, $fn=fn);
        translate([0,0,ht-lockshafthole_th]) cylinder(h=lockshafthole_th, d1=lockshafthole_db2, d2=lockshafthole_dt, $fn=fn);
        translate([0,0,ht-lockgapdepth-r]) cylinder(h=lockgapdepth-lockshafthole_th+r+0.01, d1=lockshafthole_db1, d2=lockshafthole_db2, $fn=fn);

        translate([0,0,bot]) rotate([0,90,0]) translate([0,0,-l/2]) cylinder(h=l, r=r, $fn=8);
        translate([0,0,bot]) rotate([90,0,0]) translate([0,0,-l/2]) cylinder(h=l, r=r, $fn=8);
    }
}

// lockringhole -- make a hollow space for the lock ring
// parameters:
// ht: thickness of object in which the hole is to be cut
// backside: true=cut hole out from bottom surface instead of top
// thru: true=cut a hole all the way through the object
module lockringhole(ht, backside=false, thru=true) {
    h = thru ? 2*ht : minlockht+2*lineht+ringext;
    union() {
        translate([0,0,-1]) cylinder(h=h, d=lockshaftd+2*tighttol, $fn=fn);
        translate([0,0,backside?lockneckht-lineht:ht-minlockht-ringext]) lockring(0);
    }
}


// spacer -- make a bearing (a washer) of any thickness
// parameter:
// ht: thickness of spacer
module spacer(ht) {
    difference() {
        hub(0,ht,hubdia,false);
        bearing(ht);
    }
}

// key -- crankshaft fricton-lock key
// parameters:
// ht = height that key protrudes above end of vertical shaft
// tol = tolerance (if any) for subtracting a keyhole
module key(ht, tol=0) {
    d1=0.5*shaftkeywid+tol;
    d2=d1-4*lineht;
    t = 0.25*shaftdia;
    h1 = ht-2*lineht;
    n=16;
    hull() {
        //translate([d1/2,0,h1/2]) cube([d1,d1,h1],true);
        translate([0.7*t,0,0]) cylinder(h=h1, d=d1, $fn=n);
        translate([0,t,0]) cylinder(h=h1, d=d1, $fn=n);
        translate([0,-t,0]) cylinder(h=h1, d=d1, $fn=n);
        translate([-t,0,0.5*h1]) rotate([0,0,45]) cube([d1,d1,h1], true);
        //translate([d1/2,0,ht/2]) cube([d2,d2,ht],true);
        translate([0.7*t,0,0]) cylinder(h=ht, d=d2, $fn=n);
        translate([0,t,0]) cylinder(h=ht, d=d2, $fn=n);
        translate([0,-t,0]) cylinder(h=ht, d=d2, $fn=n);
        translate([-t,0,0.5*ht]) rotate([0,0,45]) cube([d2,d2,ht], true);
    }
}

// keyhole -- cut a crankshaft keyhole
// parameters:
// ht: thickness of object in which keyhole is to be cut
// tol: tolerance by which to enlarge the hole
module keyhole(ht, shaftend=2) {
    h = ht+2;
    tol = shaftend==2 ? tighttol : tightfittol;
    // cut out the keyhole
    translate([0,0,-1]) key(h, tol);
    // cut out the expansion slot
    translate([-hubdia,-linewid/2,-1]) cube([hubdia,linewid,h]);
}

// ============= assembly view ============

// when print="assembly", show up to this layer
assemblylevel = 13;
// when print="assembly", use this crankshaft angle
assemblycrank = 35;
// set to true to cause all crankcases to show in orange in cap assembly view
pressurecrank = false;
// rotate assembly around x axis; set to 90 to turn the model so it stands upright
assemblyxrotate = 0;

if (print == "kneefootassembly") {
    translate([-sd/2,0,0]) kneefoot(true);
    translate([sd/2+2*hubdia,0,0]) kneefoot(false);
}
if (print == "assembly") {
    rotate([assemblyxrotate,0,0]) wholeassembly();
}

module wholeassembly() {
    assembly(assemblylevel, assemblycrank);
    if (assemblylevel > 99) {
        for(nc=[2:ncranks]) {
            assembly(13, assemblycrank+accumangle[nc-2], nc-1);
        }
        assembly(-2, 0, ncranks); // cap
    }
}

xcolorslip = "gold"; // color for new part in exploded view
xcolorpress = "#FF7700"; // color for pressed together parts in exploded view
acolor = "silver"; // assembly normal color
ccolor = "lightcyan"; // normal color for shaft components

module kneefoot(explode=false) {
    zoffset = explode ? 3*ccht : 0;
    color(acolor) translate([sd-sg,sf/2,0]) knee();
    color(acolor) translate([0,-sf/2,0]) rotate([0,0,180]) foot();
    color(explode ? xcolorpress : acolor) translate([-sg,-sf/2,-outerboneht-hubwasher-zoffset]) rotate([0,0,90]) bone_f(1);
    color(explode ? xcolorslip: acolor) translate([-sg,-sf/2,innerboneht+outerboneht+hubwasher+zoffset]) rotate([180,0,90]) bone_f(2);
}

module assembly(alevel, crankangle, cranklevel=0) {
    startht = cranklevel*(ccht+crankspace);
    xplodht = 6*ccht;
    guideht = 10*ccht;
    guidez = 6*ccht;
    guidecolor = "magenta";
    halftol = crankshaft_layertol*lineht/2;
    l1ht = startht + (alevel == 1 ? xplodht : 0);
    l2ht = l1ht+ccht+hubwasher+halftol+(alevel==2?xplodht:0);
    l3ht = l2ht+caht+hubwasher+(alevel==3?xplodht:0);
    l4ht = l3ht+outerboneht+hubwasher+(alevel==4?xplodht:0);
    l5ht = l4ht+outerboneht+hubwasher+(alevel==5?xplodht:0);
    l6ht = l5ht+outerboneht+hubwasher+(alevel==6?xplodht:0);
    l7ht = l6ht+outerboneht+hubwasher+(alevel==7?xplodht:0);
    l8ht = l7ht+outerboneht+hubwasher+outerboneht+(alevel==8?xplodht:0);
    l9ht = l8ht+outerboneht+hubwasher+(alevel==9?xplodht:0);
    l10ht = l9ht+outerboneht+hubwasher+(alevel==10?xplodht:0);
    l11ht = l10ht+halftol+(alevel==11?xplodht:0);
    l12ht = l2ht+caht+crankshaftht+(alevel==12?xplodht:0);
    l13ht = l12ht;
    l14ht = l2ht+crankspace-hubwasher-halftol+(alevel==14?xplodht:0);
    l15ht = l14ht+ccht+hubwasher+(alevel==15?xplodht:0);
    capht = ncranks*(ccht+crankspace)+(alevel==-1?xplodht:0);

    rotation = crankangle+$t*360;
    pm1jk = [sm*cos(rotation), sm*sin(rotation)];
    pbcd_r = [sa, -sl];
    pbcd_l = [-sa, -sl];
    pbej_r = twocircles(pm1jk, sj, pbcd_r, sb, 1);
    pbej_l = twocircles(pbcd_l, sb, pm1jk, sj, 1);
    pcgk_r = twocircles(pm1jk, sk, pbcd_r, sc, -1);
    pcgk_l = twocircles(pbcd_l, sc, pm1jk, sk, -1);

    aj_r = atan2(pbej_r[1]-pm1jk[1], pbej_r[0]-pm1jk[0]);
    aj_l = atan2(pbej_l[1]-pm1jk[1], pbej_l[0]-pm1jk[0]);
    ak_r = atan2(pm1jk[1]-pcgk_r[1], pm1jk[0]-pcgk_r[0]);
    ak_l = atan2(pcgk_l[1]-pm1jk[1], pcgk_l[0]-pm1jk[0]);
    ad_r = atan2(pbej_r[1]-pbcd_r[1], pbej_r[0]-pbcd_r[0])-90+brotate;
    ad_l = atan2(pbej_l[1]-pbcd_l[1], pbej_l[0]-pbcd_l[0])-90-brotate;
    
    pdf_r = [sa+sd*cos(ad_r), -sl+sd*sin(ad_r)];
    pdf_l = [-sa-sd*cos(ad_l), -sl-sd*sin(ad_l)];
    pfg_r = twocircles(pcgk_r, sg, pdf_r, sf, -1);
    pfg_l = twocircles(pdf_l, sf, pcgk_l, sg, -1);

    ac_r = atan2(pbcd_r[1]-pcgk_r[1], pbcd_r[0]-pcgk_r[0]);
    ac_l = atan2(pbcd_l[1]-pcgk_l[1], pbcd_l[0]-pcgk_l[0]);
    ag_r = atan2(pfg_r[1]-pcgk_r[1], pfg_r[0]-pcgk_r[0]);
    ag_l = atan2(pfg_l[1]-pcgk_l[1], pfg_l[0]-pcgk_l[0]);
    af_r = atan2(pdf_r[1]-pfg_r[1], pdf_r[0]-pfg_r[0]);
    af_l = atan2(pdf_l[1]-pfg_l[1], pdf_l[0]-pfg_l[0]);

    if (startht == 0) {
        color(alevel==2 ? xcolorpress : ccolor)
        translate([0,0,-handleht-caht-hubwasher-halftol]) rotate([0,0,rotation+180]) crankshaft_end();
    }

    if (alevel < 0) { // cap
        translate([0,0,capht]) color(alevel == -1 ? xcolorslip : ccolor) crankcase(true,false);
        if (alevel == -1) guide([0,0],capht);
    }

    if (alevel >= 1) { // crankcase
        color(alevel==1 ? xcolorslip : (alevel == 14 || pressurecrank ? xcolorpress : ccolor))
        translate([0,0,l1ht]) crankcase(showspacers=false);
        if (alevel == 1) guide([0.0],l1ht);
    }
    if (alevel >= 2) { // crankshaft arm 1
        color(alevel==2 ? xcolorslip : (alevel==12 ? xcolorpress: ccolor))
        translate([0,0,l2ht]) rotate([0,0,rotation]) crankshaft(1, 0);
        if (alevel == 2) guide([0,0],l2ht);
    }
    if (alevel >= 3) { // right bones j and c with shafts
        color(alevel==3 ? xcolorslip : (alevel==8 ? xcolorpress : acolor)) {
            translate([pm1jk[0]+sj*cos(aj_r),pm1jk[1]+sj*sin(aj_r),l3ht]) rotate([0,0,180+aj_r]) bone_j(1);
            translate([pcgk_r[0],pcgk_r[1],l3ht]) rotate([0,0,ac_r]) bone_c(1);
        }
        if (alevel == 3) {
            guide(pm1jk,l3ht);
            guide(pbcd_r,l3ht);
        }
    }
    if (alevel >= 4) { // right bone k without shaft
        color(alevel==4 ? xcolorslip : (alevel==7 ? xcolorpress : acolor))
        translate([pm1jk[0]+sk*cos(ak_r+180),pm1jk[1]+sk*sin(ak_r+180),l4ht]) rotate([0,0,ak_r]) bone_k(2);
        if (alevel==4) {
            guide(pm1jk,l4ht);
            guide(pcgk_r,l4ht);
        }
    }
    if (alevel >= 5) { // right knee-and-foot assembly, left bones j and c
        color(alevel==5 ? xcolorslip : acolor) {
            translate([pfg_r[0],pfg_r[1],l5ht-outerboneht-hubwasher]) rotate([0,0,af_r]) bone_f(1);
            translate([pfg_r[0],pfg_r[1],l5ht+innerboneht+outerboneht+hubwasher]) rotate([180,0,af_r]) bone_f(2);
            translate([pbcd_r[0],pbcd_r[1],l5ht+innerboneht]) rotate([0,180,ad_r]) knee();
            translate([pcgk_r[0],pcgk_r[1],l5ht+innerboneht]) rotate([180,0,ag_r]) foot();
        }
        color(alevel==5 ? xcolorslip : (alevel==10 ? xcolorpress : acolor)) {
            translate([pm1jk[0]+sj*cos(aj_l),pm1jk[1]+sj*sin(aj_l),l5ht]) rotate([0,0,aj_l+180]) bone_j(1);
            translate([pcgk_l[0],pcgk_l[1],l5ht]) rotate([0,0,ac_l]) bone_c(1);
        }
        if (alevel == 5) {
            guide(pm1jk,l5ht);
            guide(pbcd_l,l5ht);
            guide(pbej_r,l5ht);
            guide(pbcd_r,l5ht);
            guide(pcgk_r,l5ht);
        }
    }
    if (alevel >= 6) { // left bone k with shafts
        color(alevel==6 ? xcolorslip : (alevel==9 ? xcolorpress : acolor))
        translate([pm1jk[0]+sk*cos(ak_l),pm1jk[1]+sk*sin(ak_l),l6ht]) rotate([0,0,ak_l+180]) bone_k(1);
        if (alevel == 6) {
            guide(pm1jk,l6ht);
            guide(pcgk_l,l6ht);
        }
    }
    if (alevel >= 7) { // left knee and foot assembly, right bone k with shafts
        color(alevel==7 ? xcolorslip : acolor) {
            translate([pfg_l[0],pfg_l[1],l7ht+innerboneht+outerboneht+hubwasher]) rotate([180,0,af_l]) bone_f(1);
            translate([pfg_l[0],pfg_l[1],l7ht-outerboneht-hubwasher]) rotate([0,0,af_l]) bone_f(2);
            translate([pbcd_l[0],pbcd_l[1],l7ht]) rotate([0,0,ad_l]) knee();
            translate([pcgk_l[0],pcgk_l[1],l7ht]) rotate([0,0,ag_l]) foot();
            translate([pm1jk[0]+sk*cos(ak_r+180),pm1jk[1]+sk*sin(ak_r+180),l7ht+outerboneht]) rotate([180,0,ak_r]) bone_k(1);
        }
        if (alevel == 7) {
            guide(pbej_l,l7ht);
            guide(pbcd_l,l7ht);
            guide(pcgk_l,l7ht);
            guide(pm1jk,l7ht);
            guide(pcgk_r,l7ht);
        }
    }
    if (alevel >= 8) { // right bones j and c without shafts
        color(alevel==8 ? xcolorslip : acolor) {
            translate([pm1jk[0]+sj*cos(aj_r),pm1jk[1]+sj*sin(aj_r),l8ht]) rotate([180,0,180+aj_r]) bone_j(2);
            translate([pcgk_r[0],pcgk_r[1],l8ht]) rotate([180,0,ac_r]) bone_c(2);
        }
        if (alevel == 8) {
            guide(pm1jk,l8ht);
            guide(pbej_r,l8ht);
            guide(pbcd_r,l8ht);
            guide(pcgk_r,l8ht);
        }
    }
    if (alevel >= 9) { // left bone k without shafts
        color(alevel==9 ? xcolorslip : acolor)
        translate([pm1jk[0]+sk*cos(ak_l),pm1jk[1]+sk*sin(ak_l),l9ht]) rotate([180,0,ak_l+180]) bone_k(2);
        if (alevel == 9) {
            guide(pm1jk,l9ht);
            guide(pcgk_l,l9ht);
        }
    }
    if (alevel >= 10) { // left bones c and j without shafts
        color(alevel==10 ? xcolorslip : acolor) {
            translate([pcgk_l[0],pcgk_l[1],l10ht]) rotate([180,0,ac_l]) bone_c(2);
            translate([pm1jk[0]+sj*cos(aj_l),pm1jk[1]+sj*sin(aj_l),l10ht]) rotate([180,0,aj_l+180]) bone_j(2);
        }
        if (alevel == 10) {
            guide(pm1jk,l10ht);
            guide(pbej_l,l10ht);
            guide(pbcd_l,l10ht);
            guide(pcgk_l,l10ht);
        }
    }
    if (alevel >= 11) { // crankshaft washer
        color(alevel==11 ? xcolorslip : ccolor) {
            translate([pm1jk[0],pm1jk[1],l11ht]) spacer(hubwasher);
            translate([-sa,-sl,l11ht]) spacer(hubht1);
            translate([sa,-sl,l11ht-2*(outerboneht+hubwasher)]) spacer(hubht2);
        }
        if (alevel == 11) {
            guide(pm1jk,l11ht);
            guide(pbcd_l,l11ht);
            guide(pbcd_r,l11ht);
        }
    }
    if (alevel >= 12) { // crasnkshaft arm 2
        color(alevel==12 ? xcolorslip : (alevel==15 ? xcolorpress : ccolor))
        translate([0,0,l12ht]) rotate([0,0,rotation]) crankshaft(2, cranklevel);
        if (alevel == 12) guide(pm1jk,l12ht);
    }

    // alevel 13 displays the completely assembled section
    
    if (alevel >= 14) { // next crankcase
        color(alevel==14 ? xcolorslip : ccolor)
        translate([0,0,l14ht]) crankcase(showspacers=false);
        if (alevel == 14) guide([0,0],l14ht);
    }
    if (alevel >= 15) { // next crankshaft
        color(alevel==15 ? xcolorslip : ccolor)
        translate([0,0,l15ht]) rotate([0,0,rotation+crank_angle[ncranks][1]]) crankshaft(1, 0);
        if (alevel == 15) guide([0,0],l15ht);
    }
    
    //internal module
    module guide(xy,ht) {
        color(guidecolor) translate([xy[0],xy[1],ht-guidez]) cylinder(guideht, d=1, $fn=8);
    }
}

// ============= helper functions =============

// accumulate the sum of elements in array "a" up to index "i"
function accumsum(a, i, sum=0) = // accumlated sum of array a up to index i
    i > 0 ? sum+a[i]+accumsum(a, i-1, sum) : sum+a[0];

// Bezier cubic curve

// return coordinate of Bezier curve at location t in the interval 0...1
// must be called twice, once to pass x-coordinate coefficients and again to pass y-coordinate coefficients
function bezier_coordinate(t, n0, n1, n2, n3) =
    n0*pow((1-t), 3) + 3*n1*t*pow((1-t), 2) + 3*n2*pow(t, 2) * (1-t) + n3*pow(t, 3);

// get an x,y coordinate of a cubic Bezier curve given 4 x,y control points given in the array p[][]
function bezier_point2d(t, p) = [ // t=0...1, p=array of 4 control points
    bezier_coordinate(t, p[0][0],p[1][0],p[2][0],p[3][0]),
    bezier_coordinate(t, p[0][1],p[1][1],p[2][1],p[3][1])
];

// get coordinates of triangle vertices given lengths of 3 sides a,b,c. Two of the returned vertices lie on the x axis, with one vertex being [0,0].

// return array of 3 triangle vertices given the side lengths a,b,c
function triangle_vertices(a,b,c) =
    let( v1=[0,0], v2=[c,0], x=b*((b*b+c*c-a*a)/(2*b*c)) )
        [ v1, v2, [x, sqrt(b*b-x*x)] ];

// return the vertex angle between a and b for triangle of sides a,b,c
//law of cosines: c^2 = a^2 + b^2 − 2ab cos(C) where a,b,c are sides and C is angle between a and b
function vertex_angle(a,b,c) =
    acos((a*a + b*b - c*c)/(2*a*b));

// Solve for intersection of two circles, from http://www.ambrsoft.com/TrigoCalc/Circles2/circle2intersection/CircleCircleIntersection.htm
// As long as the left_origin is actually to the left of right_origin, ypick>0 causes the intersection with the maximum y value to be returned, otherwise the intersection with the minimum y value is returend.
function twocircles(left_origin, r1, right_origin, r2, ypick=1) =
    let(
        x1 = left_origin[0], y1 = left_origin[1],
        x2 = right_origin[0], y2 = right_origin[1],
        D2 = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2),
        D = sqrt(D2),
        a1 = D + r1 + r2,
        a2 = D + r1 - r2,
        a3 = D - r1 + r2,
        a4 = -D + r1 + r2,
        area = sqrt(a1*a2*a3*a4)/4,
        vx1 = (x1+x2)/2 + (x2-x1)*(r1*r1-r2*r2) / (2*D2),
        vx2 = 2 * (y1-y2) * area / D2,
        vy1 = (y1+y2)/2 + (y2-y1)*(r1*r1-r2*r2) / (2*D2),
        vy2 = 2 * (x1-x2) * area / D2)
    [ypick>0 ? vx1+vx2 : vx1-vx2, ypick>0 ? vy1-vy2 : vy1+vy2];
